/***********************************************************************\
* NLED Text Editor Copyright 1987-2003, Evan Alexander Weaver           *
*                                                                       *
* This source code, and the associated executable programs obtained by  *
* compiling it, are published under the GNU General Public License. See *
* the accompanying file "nled.txt" for the details of this license and  *
* some other comments about NLED and this source code.                  *
*                                                                       *
* File: NLEDATA.C                         Date Last Modified: 08-Jul-96 *
*                                                                       *
* Virtual string management. Modules that allow for creation and main-  *
* tenance of a dynamically allocated string limited in size only by     *
* memory. A doubly-linked list of nodes is used to contain the string.  *
* Any function whose name begins with 'dat' may be called by the user   *
* program.                                                              *
*                                                                       *
* Externally callable functions are:                                    *
*   datclear() - empties and reinitializes data string                  *
*   datclose() - empties data string and frees all space used           *
*   datdeln() - deletes n characters from a given position              *
*   datfappend() - appends data from an open file to data string        *
*   datfind() - locates pointer to a given position in data             *
*   datfindseq() - locates sequential access pointer to given position  *
*   datfwrite() - writes entire string to open file                     *
*   datget() - gets character at a given position                       *
*   datgetseq() - gets next character using sequential access pointer   *
*   datinsn() - inserts character n times at given position             *
*   datinss() - inserts string at given position                        *
*   datllen() - finds length of line at given position                  *
*   datmove() - moves given block to given position                     *
*   datopen() - intializes virtual string for use                       *
*   datpllen() - finds length of line prvious to given position         *
*   datprint() - displays n characters from given position              *
*   datset() - sets given position to given character                   *
*                                                                       *
* 05-Oct-91: ANSI function headers instead of "old-style"               *
* 12-Feb-92: UNIX standardization                                       *
* 08-Jul-96: Use all required standard header file                      *
\***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#if ZORTECH
    #include <dos.h>
#endif

#include "nled.h"
#include "nledkeys.h"
#include "nledfdef.h"

/* Max index of char portion of one node (baked so one node = 1024 bytes) */
#define MAX_STR (1023 - 2 - sizeof(int) - 2*sizeof(char *))

/* One node of the doubly linked list */
typedef struct noderec {
    struct noderec *prev, *next;
    int len;
    char str[MAX_STR + 1];
} *NODE;

static NODE beg_node = NULL;    /* first node in list   */
static NODE end_node = NULL;    /* last node in list    */
static NODE cur_node = NULL;    /* current node in list */
static NODE seq_node = NULL;    /* current node for special sequential access */
static INDEX cur_noff = 0;      /* offset in data of start of cur_node  */
static INDEX seq_noff = 0;      /* offset in data of start of seq_node  */
static int cur_nind = 0;        /* current index in cur_node            */
static int seq_nind = 0;        /* current index in cur_node            */
static int n_nodes = 0;         /* number of nodes      */

/***********************************************************************\
* Frees node, updating cur_nxxx vars if necessary.                      *
\***********************************************************************/
static void freenode(NODE node)
{
    if (node == NULL) return;

    if (cur_node == node) {
        cur_node = node->prev;
        if (cur_node != NULL) {
            cur_noff -= cur_node->len;
            cur_nind = 0;
        }
        else
            cur_noff = cur_nind = 0;
    }

    if (node->prev == NULL)
        beg_node = node->next;
    else
        node->prev->next = node->next;
    if (node->next == NULL)
        end_node = node->prev;
    else
        node->next->prev = node->prev;
    free (node);

    n_nodes--;
}

/***********************************************************************\
* Allocates and attaches new empty node after 'node'. Returns ERR if    *
* there was a problem (ie. no room), else returns OK.                   *
\***********************************************************************/
static int newnode(NODE node)
{
    NODE new;

#ifdef SHAREWARE
    /* for shareware version, limit # of bytes to 32k */
    if (n_nodes >= 32766/(1+MAX_STR))
        return (ERR);
#endif

    if (NULL == (new = (NODE)malloc(sizeof(struct noderec))))
        return (ERR);

    new->prev = node;
    new->len = 0;
    if (node == NULL) {
        new->next = beg_node;
        beg_node = new;
    }
    else {
        new->next = node->next;
        node->next = new;
    }
    if (new->next == NULL)
        end_node = new;
    else
        new->next->prev = new;

    n_nodes++;
    return(OK);
}

/***********************************************************************\
* Attempts to compress data starting at node 'snode', enough to free at *
* least one node, by using up empty space at the end of each node.      *
* Returns ERR if it could not be done, else returns OK.                 *
* 10-Feb-88: removed maxfill/max_fill. Nodes now filled to the brim     *
\***********************************************************************/
static int nodecompress(NODE snode)
{
    INDEX total,pos;
    int ncount,len,retcode = ERR;
    NODE inode;

    total = 0;
    inode = snode;
    ncount = -1;
    while (inode != NULL) {
        total += inode->len;
        ncount++;
        inode = inode->next;
    }
    if ((ncount < 1) || (total > (INDEX)ncount * (MAX_STR + 1))) return (ERR);

    /* mark current position in file BEFORE displaying message  */
    /* (status line resets position to current screen position) */
    pos = cur_noff + cur_nind;

    /* warn user since this can take some time */
    dispmess(" Memory is low: compressing data...", 35, TRUE);

    inode = snode;
    while (inode != NULL) {
        while (inode->len <= MAX_STR) {
            if (NULL == inode->next) break;
            len = min(inode->next->len, MAX_STR + 1 - inode->len);
            insert(inode->next->str, len, inode->len, inode->str,
             &(inode->len),MAX_STR+1);
            if (len == inode->next->len) {
                freenode(inode->next);
                retcode = OK;
            }
            else
                delete(len,0,inode->next->str,&(inode->next->len));
        }
        inode = inode->next;
    }

    /* Now, find out where we were, in the new scheme of things */
    cur_node = beg_node;
    cur_nind = 0;
    cur_noff = 0;
    datfind(pos);
    return(retcode);
}

/***********************************************************************\
* Split cur_node into two nodes at cur_nind, if cur_nind is not already *
* at the start of a node. Returns ERR if it could not get a new node in *
* which to place the second half, else returns OK.                      *
\***********************************************************************/
static int nodesplit(void)
{
    if (cur_nind != 0) {
        if (newnode(cur_node)) return (ERR);
        insert(&(cur_node->str[cur_nind]), cur_node->len-cur_nind, 0,
         cur_node->next->str, &(cur_node->next->len), MAX_STR + 1);
        delete(cur_node->len - cur_nind, cur_nind, cur_node->str,
         &(cur_node->len));
        cur_noff += cur_node->len;
        cur_node = cur_node->next;
        cur_nind = 0;
    }
    return (OK);
}

/***********************************************************************\
* Erases all data and starts afresh. Returns ERR if it could not re-    *
* initialize the data, else returns OK.                                 *
\***********************************************************************/
int datclear(void)
{
    datclose();
    return (datopen());
}

/***********************************************************************\
* Frees all nodes.                                                      *
\***********************************************************************/
void datclose(void)
{
    while (beg_node != NULL) freenode(beg_node);
}

/***********************************************************************\
* Deletes 'count' characters starting at position 'pos'. Returns the    *
* number of characters actually deleted.                                *
* 01-APR-87: Removed unused variable "tnode"                            *
\***********************************************************************/
INDEX datdeln(INDEX count, INDEX pos)
{
    INDEX delcount;
    int n;

    if (datfind(pos)) return((INDEX)0);

    delcount = count;

    n = min(cur_node->len - cur_nind, delcount);
    delete(n, cur_nind, cur_node->str, &(cur_node->len));
    delcount -= n;
    while ((delcount > 0) && (cur_node->next != NULL)) {
        n = min(cur_node->next->len, delcount);
        if (n == cur_node->next->len)
            freenode(cur_node->next);
        else
            delete(n, 0, cur_node->next->str, &(cur_node->next->len));
        delcount -= n;
    }
    if ((0 == cur_node->len) && (1 < n_nodes))
        freenode(cur_node);
    return(count - delcount);
}

/***********************************************************************\
* Appends opened file '*fp' to end of data. '*plength' and '*pnolines'  *
* are the calling program's count of characters and number of lines in  *
* the data, and are updated appropriately. 'col' is the display column  *
* of the start of the append to be used for calcualting tab expansion.  *
* 'tabsize' is the size of a tabstop, or -1 if tabs are not to be       *
* expanded. Returns ERR if it ran out of room, else returns OK.         *
* 10-Feb-88: replaced max_fill with MAX_STR. Nodes now filled to brim   *
\***********************************************************************/
int datfappend(FILE *fp, INDEX *plength, INDEX *pnolines, INDEX col,
 int tabsize)
{
    NODE inode;
    int c, retcode = OK, tabexp;

    tabexp = (tabsize >= 0) ? TRUE: FALSE;
    while (EOF != (c = fgetc(fp))) {
        if (c == LF) {
            (*pnolines)++;
            col = 0;
        }
        if ((c == TAB) && tabexp)
            do {
                if (end_node->len > MAX_STR) {
                    if (newnode(end_node)) {
                        if (nodecompress(beg_node))
                            goto error;
                        newnode(end_node);
                    }
                }
                end_node->str[end_node->len++] = ' ';
            } while ((col++) % tabsize);
        else {
            if (end_node->len > MAX_STR) {
                if (newnode(end_node)) {
                    if (nodecompress(beg_node))
                        goto error;
                    newnode(end_node);
                }
            }
            end_node->str[end_node->len++] = c;
            col++;
        }
    }
    goto ok;

error:
    retcode = ERR;
ok:
    inode = beg_node;
    *plength = 0;
    while (inode != NULL) {
        (*plength) += inode->len;
        inode = inode->next;
    }
    return(retcode);
}

/***********************************************************************\
* Locates cur_nxxx variables at position 'pos' in the data. Returns ERR *
* and positions cur_nxxx after end of last node if 'pos' is past end of *
* data, else returns OK.                                                *
\***********************************************************************/
int datfind(INDEX pos)
{
    if (cur_node == NULL) {
        cur_node = beg_node;
        cur_noff = 0;
    }

    while (pos < cur_noff) {
        cur_node = cur_node->prev;
        cur_noff -= cur_node->len;
    }
    while (pos >= cur_noff + cur_node->len) {
        if (NULL == cur_node->next) {
            cur_nind = cur_node->len;
            return (ERR);
        }
        cur_noff += cur_node->len;
        cur_node = cur_node->next;
    }
    cur_nind = pos - cur_noff;
    return (OK);
}

/***********************************************************************\
* Positions sequential data variables (seq_nxxx) at position 'pos' in   *
* the data. Returns ERR and sets seq_node to NULL if 'pos' is past end  *
* of data, else returns OK.                                             *
\***********************************************************************/
int datfindseq(INDEX pos)
{
    if (cur_node == NULL) {
        seq_node = beg_node;
        seq_noff = 0;
    }
    else {
        seq_node = cur_node;
        seq_noff = cur_noff;
    }

    while (pos < seq_noff) {
        seq_node = seq_node->prev;
        seq_noff -= seq_node->len;
    }
    while (pos >= seq_noff + seq_node->len) {
        seq_noff += seq_node->len;
        if (NULL == (seq_node = seq_node->next)) {
            seq_nind = 0;
            return (ERR);
        }
    }
    seq_nind = pos - seq_noff;
    return (OK);
}

/***********************************************************************\
* Writes all data to open file '*fp'. Returns ERR if there was an error *
* else returns OK. Leading spaces get compressed to tabs if 'tabsize'   *
* is bigger than 1 in text mode. 'accessmode' is 0 for binary, 1 for    *
* normal text, and 2 for text with Ctrl/Z at the end.                   *
* 27-Dec-88: unsigned char for Zortech C                                *
* 04-Apr-94: Added Tab compression                                      *
\***********************************************************************/
int datfwrite(FILE *fp, int tabsize, int accessmode)
{
    NODE inode;
    int i, c=LF, spcnt=0, leading=TRUE;

    /* no tab compression if each tab is only one space or binary access */
    if (tabsize == 1 || !accessmode)
        tabsize = 0;

    inode = beg_node;
    while (inode != NULL) {
        for (i = 0; i < inode->len; i++) {
            c = (unsigned char)inode->str[i];
            if (tabsize) { /* we are compressing tabs */
                if (leading) { /* we dealing with the beginning of a line */
                    if (c == ' ') {
                        spcnt++;
                        if (0 == spcnt % tabsize) { /* we hit a tab-stop */
                            c = TAB;
                            spcnt = 0; /* cancel pending spaces */
                        }
                    }
                    else if (c == TAB) /* a TAB supercedes pending spaces */
                        spcnt = 0;
                    else { /* neither a space nor a tab */
                        leading = FALSE;
                        /* write pending spaces: they can't be compressed */
                        while (spcnt) {
                            if (EOF == fputc(' ', fp))
                                return (ERR);
                            spcnt--;
                        }
                    }
                }
                if (c == LF || c == CR)
                    leading = TRUE;
            }
            if (!spcnt) /* write only if no spaces are pending */
                if (EOF == fputc(c,fp))
                    return (ERR);
        }
        inode = inode->next;
    }
    while (spcnt) { /* file may have ended with pending spaces */
        if (EOF == fputc(' ', fp))
            return (ERR);
        spcnt--;
    }
#if UNIX
    /* in UNIX text modes, make sure last line ends with a new-line */
    if (accessmode && c != LF)
        if (EOF == fputc(LF, fp))
            return (ERR);
#endif
    /* write a Control-Z if we are in Ctrl/Z text mode and need one */
    if (accessmode == 2 && c != '\x1A')
        if (EOF == fputc('\x1A', fp))
            return (ERR);
    return(OK);
}

/***********************************************************************\
* Returns character at position 'pos', or EOF if 'pos' is past end.     *
*                                                                       *
* 26-Sep-88: Zero out high byte of character                            *
* 30-Dec-88: Unsigned char instead of zeroing out high byte             *                                                         *
\***********************************************************************/
int datget(INDEX pos)
{
    if (datfind(pos)) return (EOF);

    return ((unsigned char)cur_node->str[cur_nind]);
}

/***********************************************************************\
* Returns character at current sequential read point and advances       *
* seq_nxxx variables forward one character. Returns EOF if past end of  *
* data. NOTE: datfindseq() MUST BE CALLED BEFORE THIS IS CALLED IF ANY  *
* CHANGES HAVE BEEN MADE TO THE STRUCTURE OF THE DATA SINCE THE LAST    *
* TIME datseqfind() WAS CALLED. THIS IS THE USER'S RESPONSIBILITY!!!    *
*                                                                       *
* 26-Sep-88: Zero out high byte, for Zortech C                          *
* 30-Dec-88: Unsigned char instead of zeroing out high byte             *                                                         *
\***********************************************************************/
int datgetseq(void)
{
    int c;

    if (NULL == seq_node) return (EOF);

    c = (unsigned char)(seq_node->str[seq_nind++]);
    while (seq_nind >= seq_node->len) {
        seq_noff += seq_node->len;
        seq_nind = 0;
        if (NULL == (seq_node = seq_node->next)) break;
    }
    return (c);
}

/***********************************************************************\
* Inserts character 'c' 'count' times into data at position 'pos'.      *
* Returns the number of times the character was actually inserted.      *
* 10-Feb-88: replaced max_fill with MAX_STR. Nodes now filled to brim   *
\***********************************************************************/
INDEX datinsn(char c, INDEX count, INDEX pos)
{
    int i, j;
    INDEX icount;

    if (count <= 0) return ((INDEX)0);

    if (datfind(pos)) {
        /* append it to end of data, much easier than inserting! */
        for (icount = 0; icount < count; icount++) {
            if (end_node->len > MAX_STR) {
                if (newnode(end_node)) {
                    /* we need more room */
                    if (nodecompress(beg_node)) {
                        /* can't get more room, give up */
                        count = icount;
                        goto lerror;
                    }
                    newnode(end_node);
                }
            }
            end_node->str[end_node->len++] = c;
        }
    }
    else {
        /* insert it in a non-structured but efficient manner */
        icount = count;
        while (icount > 0) {
linsert:
            /* find out how much we can insert in this node */
            i = (1 + MAX_STR) - cur_nind;
            if (i > icount) i = icount;
            if (cur_node->len + i > MAX_STR + 1) {
                /* find out how much of this node we must slide into next */
                j = cur_node->len + i - MAX_STR - 1;
                if ((cur_node->next == NULL) ||
                 (cur_node->next->len + j > MAX_STR)) {
                    if (newnode(cur_node)) {
                        /* not enough room, so make more */
                        if (nodecompress(cur_node->next)) {
                            if (nodecompress(beg_node)) {
                                /* no more room, so give up */
                                count -= icount;
                                goto lerror;
                            }
                            /* we had to compress from before current   */
                            /* point, so go figure out how much we can  */
                            /* insert again                             */
                            else
                                goto linsert;
                        }
                        else
                            newnode(cur_node);
                    }
                }
                /* move tail end of this node into next node */
                insert(&(cur_node->str[cur_node->len - j]), j, 0,
                 cur_node->next->str, &(cur_node->next->len), MAX_STR + 1);
                cur_node->len -= j;
            }
            /* insert character into this node and move on */
            insertc(c,i,cur_nind,cur_node->str,&(cur_node->len),MAX_STR + 1);
            pos += i;
            icount -= i;
            datfind(pos);
        }
    }
    return (count);

lerror:
    dispmess(" Not enough room to insert",-1,FALSE);
    beep(2);
    return(count);
}

/***********************************************************************\
* Inserts 'len' characters, starting with '*str', at position 'pos' in  *
* data. Returns the number of characters actually inserted.             *
\***********************************************************************/
int datinss(char *str, int len, INDEX pos)
{
    int i;

    if (len <= 0) return (0);

    len = datinsn(str[0],(INDEX)len,pos);

    for (i = 1; i < len; i++)
        datset(str[i],++pos);

    return (len);
}

/***********************************************************************\
* Finds and returns the length of the line in data starting at position *
* 'offset'.                                                             *
\***********************************************************************/
INDEX datllen(INDEX offset)
{
    INDEX i;
    int j;
    NODE tnode;

    if (datfind(offset)) return ((INDEX)0);

    tnode = cur_node;
    i = -(j = cur_nind);

    while (TRUE) {
        if (LF == tnode->str[j]) break;
        j++;
        while (j >= tnode->len) {
            i += j;
            if (NULL == (tnode = tnode->next)) return(i);
            j = 0;
        }
    }
    return (i + j);
}

/***********************************************************************\
* Moves the block starting at 'start' and ending just before 'end' to   *
* the position 'to'.                                                    *
* 01-APR-87: Removed unused variable "ind"                              *
\***********************************************************************/
void datmove(INDEX start, INDEX end, INDEX to)
{
    NODE snode, enode, tnode, snp, enp, tnp;
    int i, n, firstry = TRUE;
    INDEX count, index;

    if ((end <= start) || ((start <= to) && (to <= end))) return;

L200:
    datfind(to);

    if (MAX_STR + 1 - cur_node->len >= end - start) {
        /* we have room to move it directly into "to's" node */
        n = end - start;
        /* slide from "to" on to the right by n */
        for (i = cur_node->len - 1; i >= cur_nind; i--)
            cur_node->str[i + n] = cur_node->str[i];
        cur_node->len += n;
        if (to < start) start += n;

        /* copy it to new position */
        datfindseq(start);
        for (i = 0; i < n; i++)
            cur_node->str[cur_nind + i] = datgetseq();

        /* delete it from old position */
        datdeln(n, start);
    }
    else {
        /* split it up and reconnect nodes */
        if (nodesplit()) goto L250;
        tnode = cur_node;
        datfind(end);
        if (nodesplit()) goto L250;
        enode = cur_node;
        datfind(start);
        if (nodesplit()) goto L250;
        snode = cur_node;
        if (to < start) datfind(to);
        snp = snode->prev;
        enp = enode->prev;
        tnp = tnode->prev;
        snode->prev = tnp;
        enode->prev = snp;
        tnode->prev = enp;
        if (snp != NULL) snp->next = enode;
        if (enp != NULL) enp->next = tnode;
        if (tnp != NULL) tnp->next = snode;
        if (to < start) {
            cur_node = snode;
            if (tnode == beg_node) beg_node = snode;
        }
        else {
            cur_node = enode;
            if (snode == beg_node) beg_node = enode;
        }
        cur_nind = 0;
    }
    return;

    L250:
        /* No room to split it up. Try compressing file and trying once more */
        if (firstry) {
            nodecompress(beg_node);
            firstry = FALSE;
            goto L200;
        }

        /* compressing didn't do much good, so do a slow, sure method */
        dispmess(" Memory is low, move may take a while...",40,TRUE);
        for (count = 0; count < end - start; count++) {
            if (to < start) {
                i = datget(end - 1);
                for (index = end - 1; index > to; index--)
                    datset(datget(index - 1),index);
                datset(i, to);
            }
            else {
                i = datget(start);
                for (index = start; index < to - 1; index++)
                    datset(datget(index + 1),index);
                datset(i, to - 1);
            }
        }
        return;
}

/***********************************************************************\
* Open data for use by getting a node and initializing things. Returns  *
* ERR if it was already open, or if no room, else returns OK.           *
* 10-Feb-88: removed reference to max_fill. Nodes now filled to brim    *
\***********************************************************************/
int datopen(void)
{
    if (beg_node != NULL) return (ERR);
    if (newnode(beg_node)) return (ERR);
    return(OK);
}

/***********************************************************************\
* Finds and returns the length of the line in data previous to the      *
* line starting at position 'offset'.                                   *
\***********************************************************************/
INDEX datpllen(INDEX offset)
{
    INDEX i;
    int j;
    NODE tnode;

    offset -= 2;
    if ((offset < 0) || datfind(offset)) return ((INDEX)0);

    tnode = cur_node;
    i = (j = cur_nind) - tnode->len ;

    while (TRUE) {
        if (LF == tnode->str[j]) break;
        j--;
        while (j < 0) {
            i += tnode->len;
            if (NULL == (tnode = tnode->prev)) return (i + 1);
            j = tnode->len - 1;
        }
    }
    return (i + tnode->len - j);
}

/***********************************************************************\
* Displays 'len' characters starting at position 'pos'.                 *
* 16-Apr-87: Use qd_puts instead of qd_putc                             *
\***********************************************************************/
void datprint(INDEX pos, int len)
{
    int n;

    while (len > 0) {
        datfind(pos);
        n = min(cur_node->len - cur_nind, len);
        qd_puts(n, &(cur_node->str[cur_nind]));
        len -= n;
        pos += n;
    }
}

#ifdef COMMENTOUT
./***********************************************************************\
.* Displays 'len' characters starting at position 'pos', cleanly.        *
.\***********************************************************************/
.void datprintslow(INDEX pos, int len)
.{
.    int i, n, imax;
.
.    while (len > 0) {
.        datfind(pos);
.        imax = cur_nind + (n = min(cur_node->len - cur_nind, len));
.        for (i = cur_nind; i < imax; i++) qd_putc(cur_node->str[i]);
.        len -= n;
.        pos += n;
.    }
.}
#endif

/***********************************************************************\
* Sets character in position 'pos' to 'c'. Returns ERR if 'pos' is past *
* end of data, else returns OK.                                         *
\***********************************************************************/
int datset(char c,INDEX pos)
{
    if (datfind(pos)) return (ERR);

    cur_node->str[cur_nind] = c;
    return (OK);
}

#ifdef COMMENTOUT
./* Interactive display of all nodes: Used only for debugging purposes.   */
./* To use just put "datdump();" in code where you want data to be shown. */
.void datdump(void)
.{
.    NODE inode;
.    INDEX count=0;
.    int i,j,r;
.    char message[49];
.
.    sprintf(message," %d nodes: <Esc> to cancel, other to see", n_nodes);
.    dispmess(message, strlen(message), FALSE);
.    if (ESC == inputc()) return;
.
.    inode = beg_node;
.    i = 1;
.    do {
.        qd_move(0,0);
.        printstr("Node:"); printindex((INDEX)i,3);
.        printstr(" Offs:"); printindex(count,6);
.        printstr(" Len:"); printindex((INDEX)inode->len,4);
.        printstr(" Node:"); printindex(_farptr_tolong(inode),6);
.        printstr(" Prev:"); printindex(_farptr_tolong(inode->prev),6);
.        printstr(" Next:"); printindex(_farptr_tolong(inode->next),6);
.        qd_eeol();
.        qd_move(1,0);
.        printstr(" beg_node:"); printindex(_farptr_tolong(beg_node),6);
.        printstr(" end_node:"); printindex(_farptr_tolong(end_node),6);
.        printstr(" cur_node:"); printindex(_farptr_tolong(cur_node),6);
.        printstr(" cur_noff:"); printindex((INDEX)cur_noff,6);
.        printstr(" cur_nind:"); printindex((INDEX)cur_nind,6);
.        qd_eeol();
.        r = 2;
.        for (j = 0; j < inode->len; j++) {
.            if (0 == j % 80) qd_move(r++,0);
.            qd_puts(1,&(inode->str[j]));
.        }
.        qd_eeol();
.        while (r < num_rows - 1) {
.            qd_move(r++, 0);
.            qd_eeol();
.        }
.        if (ESC == inputc()) break;
.        count += inode->len;
.        inode = inode->next;
.        i++;
.    } while (NULL != inode);
.    setdrawall();
.}
#endif
/*--------------------------END-OF-NLEDATA.C---------------------------*/
