/***********************************************************************\
* NLED Text Editor Copyright 1987-2003, Evan Alexander Weaver           *
*                                                                       *
* This source code, and the associated executable programs obtained by  *
* compiling it, are published under the GNU General Public License. See *
* the accompanying file "nled.txt" for the details of this license and  *
* some other comments about NLED and this source code.                  *
*                                                                       *
* File: NLEDIO.C                          Date Last Modified: 15-Feb-02 *
*                                                                       *
* Input/Output modules for NLED (Neat Little EDitor).                   *
* Functions are:                                                        *
*   beep() - beeps if beep level is low enough                          *
*   fgetnb() - gets specified number of bytes from a file               *
*   fputnb() - writes specified number of bytes to a file               *
*   inputc() - gets next character from active macro or keyboard        *
*   inputs() - input a string from the command line                     *
*   insfile() - insert a file at current location                       *
*   loadcfg() - read configuration file                                 *
*   namefile() - give new name to current file                          *
*   newfile() - loads new file into NLED                                *
*   printhex() - prints int as 4 hex digits                             *
*   printindex() - prints an INDEX in given space (uses * if overflow)  *
*   printsstr() - prints substring of a string                          *
*   printstr() - prints null-terminated string                          *
*   readfile() - reads file into NLED                                   *
*   savecfg() - saves current configuration                             *
*   savefile() - saves current file                                     *
*   saveyn() - asks if file should be saved, saving it if yes           *
*                                                                       *
* 05-Oct-91: ANSI function headers instead of "old-style"               *
* 12-Feb-92: UNIX standardization                                       *
* 08-Jul-96: Win95 port                                                 *
* 03-Feb-02: Release 2.5 cleanup                                        *
* 15-Feb-02: Increased file name array sizes to MAX_FILE                *
\***********************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "nled.h"
#include "nledkeys.h"
#include "nledvdef.h"
#include "nledfdef.h"

/***********************************************************************\
* Sounds a beep at the console.                                         *
* 12-Feb-92: use qd_beep() instead of bdos                              *
\***********************************************************************/
void beep(int n)
{
    if (beep_level >= n)
        qd_beep();
}

/***********************************************************************\
* Reads 'count' bytes into 'pointer[]' from open file '*fp'. Returns    *
* ERR if there was a problem, OK otherwise.                             *
\***********************************************************************/
int fgetnb(int count, char *pointer, FILE *fp)
{
    int c,i;

    for (i = 0; i < count; i++) {
        if (EOF == (c = fgetc(fp))) return (EOF);
        pointer[i] = (char)c;
    }
    return (OK);
}

/***********************************************************************\
* Writes 'count' bytes, starting at 'pointer', to open file '*fp'.      *
* Returns EOF if error, OK otherwise.                                   *
\***********************************************************************/
int fputnb(int count, char *pointer, FILE *fp)
{
    int i;

    for (i = 0; i < count; i++)
        if (EOF == fputc((unsigned char)(pointer[i]), fp)) return (EOF);
    return(OK);
}

/***********************************************************************\
* Gets a character from the keyboard, waiting if necessary. Returns an  *
* integer with high byte=0 and low byte=ASCII code for a normal ASCII   *
* character. For an IBM extended code key, high byte=0xFF and low byte= *
* ASCII code. Note that for an extended key the value returned is <0.   *
* If a function key macros is being processed, it returns the next code *
* in the key's sequence instead of going to the keyboard.               *
* 21-Sep-88: Zero out high-order byte of bdos() return for Zortech C    *
* 12-Feb-92: Use qd_getch() instead of bdos() for input                 *
\***********************************************************************/
int inputc(void)
{
    /* see if a function key has been called for */
    if (func_key >= 0) {
        if (func_ind < func_mac[func_key][0]) {
            msg_flag = FALSE;
            return(func_mac[func_key][++func_ind]);
        }
        else
            func_key = -1;
    }

    /* If we're still here, we need to get from keyboard */

    /* update display with latest changes first of all */
    if (status_flag) {
        dispstatus();
        msg_flag = FALSE;
    }
    else qd_flush();

    return (qd_getch());
}

/***********************************************************************\
* Displays 'prompt[]' (null terminated) and 'string[]' (length '*plen') *
* on the status line and allows user to edit 'string[]' (max length is  *
* 'maxlen'). Returns ESCHIT if user escapes, else length of new string. *
* 05-Apr-94: key alternatives for UNIX                                  *
* 03-Feb-02: allow editing of strings longer than available display     *
*            width, which is MAX_PM - strlen(prompt)                    *
\***********************************************************************/
int inputs(char *string, int *plen, int maxlen, char *prompt)
{
    char output[MAX_PM + 1];
    int col, c, flag = TRUE, spos = 0, offs, outlen, displen;

    strcpy(output, prompt);
    col = strlen(prompt);
    displen = MAX_PM - col;
    offs = 0; /* min(*plen, MAX_PM - col); */

    while (flag) {
        outlen = col;
        if (*plen) 
         insert(string + spos, *plen - spos, col, output, &outlen, MAX_PM);
        dispnmess(output, outlen, (col + offs), FALSE);
        c = inputc();
        switch (c) {
        case ESC:
            if (*plen == 0) return(ESCHIT);

            offs = spos = *plen = 0;
            string[0] = '\0';
            break;

        case LF:
        case CR:
            flag = FALSE;
            break;

        case HOME:
#ifdef ALT_HOME
        case ALT_HOME:
#endif
            offs = spos = 0;
            break;

        case END:
#ifdef ALT_END
        case ALT_END:
#endif
            if (*plen > displen) {
                spos = *plen - displen;
                offs = displen;
            }
            else {
                spos = 0;
                offs = *plen;
            }
            break;

        case INS:
#ifdef ALT_INS
        case ALT_INS:
        case ALT_INS1:
#endif
            ins_mode = !ins_mode;
            break;

        case BS:
            if (spos + offs <= 0) break;
            if (offs > 1 || spos == 0)
                offs--;
            else
                spos--;
        case DEL:
            if (spos + offs < *plen)
                delete(1, spos + offs, string, plen);
            break;

        case LEFT:
            if (spos + offs > 0) {
                if (offs > 1 || spos == 0)
                    offs--;
                else
                    spos--;
            }
            break;

        case RGHT:
            if (spos + offs < *plen) {
                if (offs < displen - 1 || spos == *plen - displen)
                    offs++;
                else
                    spos++;
            }
            break;

        default:
            if ((spos + offs < maxlen) && (!ins_mode || *plen < maxlen) &&
             ((c == HEX) || ((' ' <= c) && (c <= LASTPCHAR)))) {
                if (c == HEX)
                    if ((c = inputhex()) < 0) break;
                if ((ins_mode) || (spos + offs == *plen)) {
                    insertc(c,1,spos+offs,string,plen,maxlen);
                }
                else {
                    string[spos + offs] = c;
                }
                if (offs < displen - 1 || spos == *plen - displen)
                    offs++;
                else
                    spos++;
            }
            else beep(1);
            break;
        }
    }
    return (*plen);
}

/***********************************************************************\
* Asks for name of file to import, and imports it to current position.  *
* Returns ESCHIT if user cancels operation, else OK.                    *
* 01-APR-87: Added message if file can't be read                        *
* 02-Feb-02: Allow longer file name                                     *
* 15-Feb-02: use MAX_FILE for file name length, made ifile static       *
\***********************************************************************/
int insfile(void)
{
    INDEX lends,n;
    static char ifile[MAX_FILE+1]="";
    char message[50];
    int len;

    len = strlen(ifile);
    if (ESCHIT == inputs(ifile,&len,MAX_FILE," Import file:"))
        return (ESCHIT);
    ifile[len] = '\0';

    filltoend();
    lends = len_ds;

    if (readfile(ifile,scr_left+col_no+1)) {
        strcpy(message," No such file:");
        strncat(message,ifile,30);
        dispmess(message,-1,FALSE);
        return(ERR);
    }

    if (0 < (n = len_ds - lends)) {
        datmove(lends, len_ds, abs_pos);
        blockupd(abs_pos, n);
        dispcalc(row_no);
        setdraw();
        draw_erow = num_rows - 2;
        edit_flag = TRUE;
    }
    return (OK);
}

/***********************************************************************\
* Load configuration from NLED.CFG file.                                *
* 17-Feb-92: case sensitivity, color, pass full pathname                *
* 04-Apr-94: Tab compression                                            *
\***********************************************************************/
int loadcfg(char *fullpath)
{
    int i;
    FILE *fp;
    if (NULL == (fp = fopen("nledcfg.bin","rb")))
        if (NULL == (fp = fopen(fullpath,"rb")))
            return (ERR);

    if (EOF == fgetnb((sizeof func_mac), (char *)func_mac, fp))
            goto error;
    for (i = 0; i < 40; i++)
        if ((0 > func_mac[i][0]) || (9 < func_mac[i][0]))
            goto error;
    tab_size = fgetc(fp);
    if ((tab_size < 0) || (tab_size > 15))
        goto error;
    beep_level = fgetc(fp);
    if ((beep_level < 0) || (beep_level > 2))
        goto error;
    if (EOF == (text_mode = fgetc(fp)))
        goto error;
    if (EOF == (ins_mode = fgetc(fp)))
        goto error;

    /* case sensitivity is new, so may not be on file */
    i = case_sens; /* store current setting just in case */
    if (EOF == (case_sens = fgetc(fp)))
        case_sens = i;

    /* color selection is new, so may not be on file */
    /* only read it if user didn't set it */
    if (color_setting == -1) {
        if (EOF == (color_setting = fgetc(fp)))
            color_setting = 0;
    }

    /* tab compression is new, so may not be on file */
    i = tab_compress; /* store current setting just in case */
    if (EOF == (tab_compress = fgetc(fp)))
        tab_compress = i;

    fclose(fp);
    return(OK);

  error:
    fclose(fp);
    return(ERR);
}

/***********************************************************************\
* Interactively renames file currently being edited.                    *
* 02-Feb-02: larger file name buffer to match main()                    *
* 15-Feb-02: use MAX_FILE for file name length, made temp name static   *
\***********************************************************************/
int namefile(char *prompt, char *defname)
{
    static char tfn[MAX_FILE+1] = "";
    int retcode,len;

    if (defname)    /* else defname is a NULL pointer: use old value in tfn */
        strcpy(tfn,defname);
    len = strlen(tfn);
    if (0 <= (retcode = inputs(tfn,&len,MAX_FILE,prompt))) {
        tfn[len] = '\0';
        strcpy(file_name,tfn);
    }
    return(retcode);
}

/***********************************************************************\
* Loads new file into memory, erasing previous one.                     *
* 01-APR-87: Added message if can't read file                           *
\***********************************************************************/
int newfile(void)
{
    int retcode;
    char message[50];

    datclear();
    nol_ds = 1;
    len_ds = 0;

    retcode = readfile(file_name,1);
    if (retcode) {
        strcpy(message," New file:");
        strncat(message,file_name,38);
        dispmess(message,-1,FALSE);
    }

    dispinit(0,0,0,0);
    edit_flag = FALSE;
    return(retcode);
}

/***********************************************************************\
* Displays 'val' as 2 hex digits if high byte is 0, else displays 'xx'. *
* 27-APR-87: Use qd_puts instead of qd_putc                             *
\***********************************************************************/
void printhex(int val)
{
    char temp,outs[2];

    if (0xFF00 & val)
        qd_puts(2,"xx");
    else {
        temp = val >> 4;
        outs[0] = (temp > 9) ? (temp + ('A' - 10)) : (temp + '0');
        temp = 0x0F & val;
        outs[1] = (temp > 9) ? (temp + ('A' - 10)) : (temp + '0');
        qd_puts(2,outs);
    }
}

/***********************************************************************\
* Displays INDEX quantity 'val' as decimal integer, left justified in a *
* field 'maxlen' characters wide, using * if there is not enough room.  *
* 27-APR-87: Use qd_puts instead of qd_putc                             *
* 12-Feb-92: use more digits                                            *
\***********************************************************************/
void printindex(INDEX val, int maxlen)
{
    char temp[15]; /* NOTE: assumes no more than 15 digits! */
    char outs[15];
    int i = 0, io = 0;

    if (val < 0) {
        outs[io++] = '-';
        val = -val;
    }
    do {
        temp[i++] = '0' + (char)(val % 10);
        val /= 10;
    } while (val != 0);

    if (i > maxlen + io)
        while (io < maxlen)
            outs[io++] = '*';
    else {
        while (i-- > 0)
            outs[io++] = temp[i];
        while (io < maxlen)
            outs[io++] = ' ';
    }
    qd_puts(maxlen,outs);
}

/***********************************************************************\
* Displays 'count' characters, starting at position 'pos' in 'string[]' *
* 16-Apr-87: Use qd_puts instead of qd_putc                             *
\***********************************************************************/
void printsstr(int pos, int count, char *string)
{
    if (count > 0)
        qd_puts(count, &(string[pos]));
}

/***********************************************************************\
* Displays null terminated character string 'string[]'.                 *
* 16-Apr-87: Use qd_puts instead of qd_putc                             *
\***********************************************************************/
void printstr(char *string)
{
    int len;

    if (0 < (len = strlen(string)))
        qd_puts(len,string);

}

/**********************************************************************\
* Appends file to end of file currently in memory.                     *
* 01-APR-87: Removed message if can't read file                        *
\**********************************************************************/
int readfile(char *name, INDEX col)
{
    FILE *fp;
    char message[50];

    if (NULL == (fp = fopen(name,(text_mode?"r":"rb"))))
        return(ERR);

    strcpy(message," Reading file:");
    strncat(message,name,29);
    dispmess(message,strlen(message),TRUE);

    if (datfappend(fp, &len_ds, &nol_ds, col, (text_mode?tab_size:-1)))
        dispmess(" File read truncated due to lack of space",-1,FALSE);
    fclose(fp);
    return(OK);
}

/***********************************************************************\
* Saves current configuration in current directory.                     *
* 01-APR-87: Remove unused variable "i"                                 *
* 14-Feb-92: case sesitivity, color                                     *
* 28-Mar-93: save color as 0 (default colors) if not set                *
* 04-Apr-94: Tab compression                                            *
\***********************************************************************/
int savecfg(void)
{
    FILE *fp;


    if (NULL == (fp = fopen("nledcfg.bin","wb"))) return(ERR);

    if ((EOF == fputnb((sizeof func_mac), (char *)func_mac, fp)) ||
     (EOF == fputc(tab_size, fp)) ||
     (EOF == fputc(beep_level, fp)) ||
     (EOF == fputc(text_mode, fp)) ||
     (EOF == fputc(ins_mode,fp)) ||
     (EOF == fputc(case_sens,fp)) ||
     (EOF == fputc((color_setting==-1) ? 0 : color_setting,fp)) ||
     (EOF == fputc(tab_compress, fp))) {
        fclose(fp);
        return(ERR);
    }

    fclose(fp);
    return (OK);
}

/***********************************************************************\
* Saves file currently being edited. Returns ERR if there was a problem *
* and OK if not.                                                        *
* 12-Oct-87: backup file message now gives backup file name             *
* 11-Mar-88: write EOF (Ctrl-Z) in text mode                            *
* 05-Oct-91: put comment in empty loop to stop compiler warning         *
* 13-Feb-92: UNIX stuff, optional Ctrl-Z                                *
* 28-Mar-93: put " in UNIX cp command, in case they put spaces in name  *
* 04-Apr-94: changed datfwrite call to enable tab compression, and      *
*            changed UNIX file opening so that the backup is copied and *
*            the original is simply overwritten (preserving ownership)  *
* 08-Jul-96: Make backup file name larger (UNIX & Win95 add to length)  *
* 15-Feb-02: use MAX_FILE for file name length                          *
\***********************************************************************/
int savefile(void)
{
    int i;
    char backup[MAX_FILE+1], message[50];
    FILE *fp;
#if UNIX
    char backcommand[2*MAX_FILE + 10];
#endif

    if (0 == strlen(file_name))
        if ((ESCHIT == namefile(" File name:",file_name)) || 
            (0 == strlen(file_name))) {
                dispmess(" No file saved",-1,FALSE);
                return(ERR);
        }

    strcpy(backup,file_name);

#if (UNIX || WIN95)
    /* in UNIX/Windows, we can just stick a .bak on the end */
    if (strlen(backup) > MAX_FILE - 4)
        backup[MAX_FILE - 4] = '\0';
    strcat(backup, ".bak");
#else
    /* in DOS, we have to stip off the old extension */
    for (i = strlen(backup)-1; (i >= 0) && (backup[i] != '.'); i--)
        /* Nothing */;
    if (i < 0)  strcat(backup,".BAK");
    else        strcpy(backup+i,".BAK");
#endif

    strcpy(message," Saving file:");
    strncat(message,file_name,30);
    dispmess(message,-1,TRUE);

    if (NULL != (fp = fopen(file_name,"r"))) {
        fclose(fp);
#if UNIX
        /* use cp to make the backup file */
        strcpy(backcommand, "cp \"");
        strcat(backcommand, file_name);
        strcat(backcommand, "\" \"");
        strcat(backcommand, backup);
        strcat(backcommand, "\"");
        system(backcommand);
#else
        /* just rename the file, deleting existing backup if necessary */
        unlink(backup);
        if (rename(file_name,backup)) {
            strcpy(message," Error making backup file:");
            strncat(message,backup,21);
            dispmess(message,-1,TRUE);
        }
#endif
    }

    if ((NULL == (fp = fopen(file_name,(text_mode?"w":"wb")))) ||
     datfwrite(fp, tab_compress ? tab_size : 0, text_mode)) {
        if (fp) fclose (fp);
        strcpy(message," Error saving file:");
        strncat(message,file_name,28);
        dispmess(message,-1,FALSE);
        return(ERR);
    }
    fclose (fp);
    strcpy(message," File saved:");
    strncat(message,file_name,29);
    dispmess(message,-1,FALSE);
    edit_flag = FALSE;
    return(OK);
}

/***********************************************************************\
* Checks to see if file has been edited since last save, and            *
* interactively controls saving according to user's desires. Returns    *
* result of the save if save is chosen or OK if no save is done.        *
\***********************************************************************/
int saveyn(void)
{
    int c, gotit = FALSE, retcode = OK;

    if (edit_flag) {
        while (!gotit) {
            dispmess(" File not saved. Save it?   (Y/N)",26,FALSE);
            beep(2);
            c = inputc();
            c = toupper(c);
            gotit = TRUE;
            switch (c) {
                case 'Y':
                    retcode = savefile();
                    break;

                case 'N':
                    retcode = OK;
                    break;

                case ESC:
                    retcode = ESCHIT;
                    break;

                default:
                    beep(1);
                    gotit = FALSE;
                    break;
            }
        }
    }
    return(retcode);
}
/*--------------------------END-OF-NLEDIO.C---------------------------*/
